from typing import TypeVar, Generic, List


T = TypeVar("T")


class Stack(Generic[T]):
    def __init__(self) -> None:
        self.__stack: List[T] = []

    def set_current(self, obj: T):
        self.__stack.append(obj)

    def is_current(self, obj: T):
        return self.get_current() is obj

    def reset_current(self):
        return self.__stack.pop()

    def __len__(self):
        return len(self.__stack)

    def get_current(
        self,
    ):
        if len(self.__stack) <= 0:
            return None

        return self.__stack[len(self.__stack) - 1]



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    List,
)

from typing_extensions import Protocol

if TYPE_CHECKING:
    from .effect import Effect


class IScope(Protocol):
    def add_effect(self, effect: Effect):
        ...

    def dispose(self):
        ...


class Scope(IScope):
    def __init__(self) -> None:
        self._effects: List[Effect] = []

    def add_effect(self, effect: Effect):
        self._effects.append(effect)

    def dispose(self):
        for effect in self._effects:
            effect.cleanup_deps()

        self._effects.clear()



from collections.abc import Iterable
import sys
from typing import Tuple, Union, TypeVar, cast, Dict, Any, Generic
from typing_extensions import SupportsIndex
from signe import batch, createSignal
from signe.utils import TGetter, TSetter
from collections import UserList, UserDict


T = TypeVar("T")

_proxy_map = {}


class Proxy:
    def __init__(self, obj) -> None:
        _proxy_map[id(obj)] = self

    def toRaw(self):
        pass


TKeySignalMapKey = Union[int, str]
TKeySignalMapValue = Tuple[TGetter[Any], TSetter[Any]]
TKeySignalMap = Dict[TKeySignalMapKey, TKeySignalMapValue]


class CollectionsProxy(Proxy):
    def __init__(self, obj) -> None:
        super().__init__(obj)
        self._keySignalMap: TKeySignalMap = {}

    def _create_signal(self, key: TKeySignalMapKey, value):
        signal = None
        if key not in self._keySignalMap:
            signal = createSignal(value,debug_name=str(key))
            self._keySignalMap[key] = signal
        else:
            signal = self._keySignalMap[key]

        return signal

    def _try_get_signal(self, key: TKeySignalMapKey):
        return self._keySignalMap.get(key, (None, None))

    def _ex_track(self, key, value):
        getter, _ = self._ex_track_return_signal(key, value)
        getter()

    def _ex_track_return_signal(self, key, value):
        getter, setter = self._create_signal(key, value)
        return getter, setter


class ListProxy(CollectionsProxy, UserList):
    def __init__(self, obj) -> None:
        super().__init__(obj)
        assert isinstance(obj, list)
        self.data = obj

    def __track_all_items(self):
        for idx in range(len(self.data)):
            self._ex_track(idx, self.data[idx])

    def __track_len(self):
        self._ex_track("__len__", len(self.data))

    def __trigger_len(self):
        _, setter = self._try_get_signal("__len__")
        if setter is not None:
            setter(len(self.data))

    def __trigger_index(self, start_idx: int):
        @batch
        def _():
            for idx in range(start_idx, len(self.data) + 1):
                _, setter = self._try_get_signal(idx)
                if setter is not None:
                    value = None if idx >= len(self.data) else self.data[idx]
                    setter(value)

    def __iter__(self):
        for item in self.data:
            yield createReactive(item)

        self.__track_all_items()
        self.__track_len()

    def __str__(self) -> str:
        value = str(self.data)  # type: ignore
        self.__track_all_items()
        return value

    def __getitem__(self, _key: int):
        value = self.data[_key]
        self._ex_track(_key, value)
        return createReactive(value)

    def __setitem__(self, _key: int, newVal):
        self.data[_key] = newVal

        _, setter = self._try_get_signal(_key)
        if setter is not None:
            setter(newVal)

    def pop(self, i: int = -1) -> Any:
        start = i
        if start == -1:
            start = len(self.data) - 1

        value = self.data.pop(i)

        @batch
        def _():
            self.__trigger_len()
            self.__trigger_index(start)

        return value

    def append(self, item: Any) -> None:
        self.data.append(item)

        @batch
        def _():
            self.__trigger_len()
            self.__trigger_index(0)

    def extend(self, other: Iterable) -> None:
        self.data.extend(other)

        @batch
        def _():
            self.__trigger_len()
            self.__trigger_index(0)

    def __len__(self) -> int:
        self.__track_len()
        return len(self.data)

    def clear(self) -> None:
        self.data.clear()

        @batch
        def _():
            self.__trigger_len()
            self.__trigger_index(0)

    def toRaw(self):
        return self.data

    def count(self, item: Any) -> int:
        if isReactive(item):
            item = item.toRaw()
        value = self.data.count(item)
        self.__track_len()
        self.__track_all_items()
        return value

    def index(
        self, item: Any, __start: SupportsIndex = 0, __stop: SupportsIndex = sys.maxsize
    ) -> int:
        value = self.data.index(item, __start, __stop)
        self.__track_len()
        self.__track_all_items()
        return value

    def insert(self, i: int, item: Any) -> None:
        self.data.insert(i, item)

        @batch
        def _():
            self.__trigger_len()
            self.__trigger_index(i)

    def remove(self, item: Any) -> None:
        org_len = len(self.data)
        self.data.remove(item)
        if len(self.data) < org_len:
            pass

            @batch
            def _():
                self.__trigger_len()
                self.__trigger_index(0)

    def sort(self, *, key: None = None, reverse: bool = False):
        self.data.sort(key=key, reverse=reverse)
        self.__trigger_index(0)

    def reverse(self):
        self.data.reverse()
        self.__trigger_index(0)


TDict = TypeVar("TDict", bound=Dict)


class DictProxy(CollectionsProxy, UserDict, Generic[TDict]):
    def __init__(self, obj: TDict) -> None:
        super().__init__(obj)
        assert isinstance(obj, dict)
        self.data = obj

    def __contains__(self, key: object) -> bool:
        result = key in self.data
        self._ex_track(key, result)
        return result

    def __track_all_items(self):
        for key, value in self.data.items():
            self._ex_track(key, value)

    def __track_len(self):
        self._ex_track("__len__", len(self.data))

    def __trigger_len(self):
        _, setter = self._try_get_signal("__len__")
        if setter is not None:
            setter(len(self.data))

    def __trigger_key(self, key, value):
        _, setter = self._try_get_signal(key)
        if setter is not None:
            setter(value)

    def __getitem__(self, _key: Any) -> Any:
        value = self.data[_key]
        self._ex_track(_key, value)
        return createReactive(value)

    def __setitem__(self, _key: Any, item: Any) -> None:
        has_key = _key in self.data
        self.data[_key] = item

        _, setter = self._try_get_signal(_key) # type: ignore
        if setter is not None:
            setter(item)

        if not has_key:
            @batch
            def _():
                self.__trigger_len()

    def __delitem__(self, key: Any) -> None:
        del self.data[key]

        @batch
        def _():
            self.__trigger_len()
            self.__trigger_key(key, None)

    def __len__(self) -> int:
        self.__track_len()
        return len(self.data)

    def get_signal(self, _key: Any):
        value = self.data[_key]
        return self._ex_track_return_signal(_key, value)

    def toRaw(self):
        return self.data


# d = DictProxy({"a": 1})


def isReactive(obj):
    return isinstance(obj, Proxy)


def createReactive(obj: T) -> T:
    if isinstance(obj, (str, int, float, tuple, bool)):
        # warn(f"Type [{type(obj)}] cannot be used as the target object of reactive.")
        return obj

    if isinstance(obj, Proxy):
        return obj

    if id(obj) in _proxy_map:
        return _proxy_map[id(obj)]

    if isinstance(obj, list):
        return cast(T, ListProxy(obj))

    if isinstance(obj, dict):
        return cast(T, DictProxy(obj))

    return obj



from __future__ import annotations
from typing import TYPE_CHECKING, Set, Dict
from .collections import Stack

from .effect import Effect

if TYPE_CHECKING:
    from .signal import Signal


class Executor:
    def __init__(self) -> None:
        self._observers: Set[Signal] = set()
        self.effect_running_stack = Stack[Effect]()
        self.execution_scheduler_stack = Stack[ExecutionScheduler]()
        self.__defalut_executionScheduler = ExecutionScheduler()

    @property
    def current_execution_scheduler(self):
        return (
            self.execution_scheduler_stack.get_current()
            or self.__defalut_executionScheduler
        )

    @property
    def is_running(self):
        return (
            self.current_execution_scheduler.is_running
            or len(self.effect_running_stack) > 0
        )


class ExecutionScheduler:
    def __init__(self) -> None:
        self.__tick = 0
        self.__signal_updates: Dict[Signal, None] = {}
        self.__effect_updates: Dict[Effect, None] = {}
        self.__running = False

    @property
    def tick(self):
        return self.__tick

    def add_signal(self, signal: Signal):
        self.__signal_updates[signal] = None
        return self

    def add_effect(self, effect: Effect):
        self.__effect_updates[effect] = None
        return self

    def next_tick(self):
        self.__tick += 1
        return self

    @property
    def is_running(self):
        return self.__running

    def run(self):
        self.__effect_updates.clear()
        self.next_tick()
        count = 0
        self.__running = True

        while len(self.__signal_updates) > 0 or len(self.__effect_updates) > 0:
            self._run_signal_updates()
            self._run_effect_updates()

            count += 1
            if count >= 10000:
                raise Exception("exceeded the maximum number of execution rounds.")

        self.__tick = 0
        self.__running = False

    def cleanup_signal_updates(self):
        self.__signal_updates.clear()

    def cleanup_effect_updates(self):
        self.__effect_updates.clear()

    def _run_signal_updates(self):
        signals = list(self.__signal_updates.keys())
        for s in signals:
            s.update()

        self.cleanup_signal_updates()

    def _run_effect_updates(self):
        effects = sorted(self.__effect_updates.keys(), key=lambda x: x.priority_level)
        for s in effects:
            s.update()
            s._reset_age()

        self.cleanup_effect_updates()


class BatchExecutionScheduler(ExecutionScheduler):
    def __init__(self) -> None:
        super().__init__()

    def run(self):
        pass

    def run_batch(self):
        return super().run()



from signe.utils import (
    effect,
    # effect_with_opts,
    createSignal,
    computed,
    batch,
    cleanup,
    on,
    scope,
)
from signe.reactive.proxy import createReactive, isReactive


__version__ = "0.2.6"



from __future__ import annotations
from typing import TYPE_CHECKING, Any, TypeVar, Set, Callable, Optional, Generic, cast
from .consts import EffectState
from itertools import chain

if TYPE_CHECKING:
    from .runtime import Executor
    from .signal import Signal

T = TypeVar("T")


class EffectOption:
    def __init__(self, level=0) -> None:
        self.level = level


class Effect(Generic[T]):
    _g_id = 0

    def __init__(
        self,
        executor: Executor,
        fn: Callable[[], T],
        debug_trigger: Optional[Callable] = None,
        priority_level=1,
        debug_name: Optional[str] = None,
        capture_parent_effect=True,
    ) -> None:
        Effect._g_id += 1
        self.id = Effect._g_id
        self.__executor = executor
        self.value: Optional[T] = None
        self.fn = fn
        self._age = 0
        self._state = EffectState.CURRENT
        self.__dep_signals: Set[Signal] = set()
        self.__priority_level = priority_level
        self.__debug_name = debug_name

        """
        When one effect is triggered by another effect, 
        the former belongs to a pre dependency 
        and the latter belongs to a next dependency.

        @effect
        def a():
            return 1

        @effect
        def b():
            return a()

        a is pre dep effect
        b is next dep effect

        b should be push to next dep of a
        a should be push to pre dep of b
        """
        self.__pre_dep_effects: Set[Effect] = set()
        self.__next_dep_effects: Set[Effect] = set()

        self._sub_effects: list[Effect] = []

        self._cleanup_callbacks: list[Callable] = []

        self._debug_trigger = debug_trigger

        """
        This method must be executed before Method __run_fn. 
        If an effect object is created 
        during the execution of a parent effect object, 
        it is considered as its child object.
        """
        if capture_parent_effect:
            self.__try_add_self_to_parent_sub_effect()

        self.__run_fn()
        self.__init_no_deps = (
            len(self.__pre_dep_effects)
            + len(self.__next_dep_effects)
            + len(self.__dep_signals)
        ) <= 0

    def __get_all_dep_effects(self):
        return chain(self.__pre_dep_effects, self.__next_dep_effects)

    def __try_add_self_to_parent_sub_effect(self):
        current_effect = self.__executor.effect_running_stack.get_current()

        if current_effect is not None and current_effect is not self:
            current_effect._add_sub_effect(self)

    @property
    def priority_level(self):
        return self.__priority_level

    def _get_pre_dep_effects(self):
        return list(self.__pre_dep_effects)

    def _get_next_dep_effects(self):
        return list(self.__next_dep_effects)

    def _add_sub_effect(self, effect: Effect):
        self._sub_effects.append(effect)

    def add_dep_signal(self, signal: Signal):
        self.__dep_signals.add(signal)

    def add_pre_dep_effect(self, effect: Effect):
        self.__pre_dep_effects.add(effect)

    def add_next_dep_effect(self, effect: Effect):
        self.__next_dep_effects.add(effect)

    def getValue(self):
        if not self.__init_no_deps:
            tick = self.__executor.current_execution_scheduler.tick
            current_effect = self.__executor.effect_running_stack.get_current()

            if current_effect:
                if self._age == tick:
                    if self._state == EffectState.RUNNING:
                        raise Exception("circular running")

                    self.update()

                self.add_next_dep_effect(current_effect)
                current_effect.add_pre_dep_effect(self)

        return cast(T, self.value)

    def _push_scheduler(self):
        tick = self.__executor.current_execution_scheduler.tick

        if self._age >= tick:
            return

        self._age = tick
        self._state = EffectState.STALE

        self.__executor.current_execution_scheduler.add_effect(self)

        for effect in self.__next_dep_effects:
            effect._push_scheduler()

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.getValue()

    def update(self):
        if self._state != EffectState.STALE:
            return

        self.__run_fn()

    def add_cleanup_callback(self, callback: Callable):
        self._cleanup_callbacks.append(callback)

    def __run_fn(self):
        self._cleanup_source_before_update()

        try:
            self.__executor.effect_running_stack.set_current(self)

            self._state = EffectState.RUNNING

            self.cleanup_deps()

            self.value = self.fn()
            if self._debug_trigger:
                self._debug_trigger()

            self._state = EffectState.CURRENT

        finally:
            self.__executor.effect_running_stack.reset_current()

    def cleanup_dep_effect(self, effect: Effect):
        if effect in self.__pre_dep_effects:
            self.__pre_dep_effects.remove(effect)

        if effect in self.__next_dep_effects:
            self.__next_dep_effects.remove(effect)

    def cleanup_deps(self):
        for s in self.__dep_signals:
            s.cleanup_dep_effect(self)

        self.__dep_signals.clear()

        for effect in self.__get_all_dep_effects():
            effect.cleanup_dep_effect(self)

        self.__pre_dep_effects.clear()
        self.__next_dep_effects.clear()

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, self.__class__):
            return self.__hash__() == __value.__hash__()

        return False

    def _cleanup_source_before_update(self):
        self._cleanup_sub_effects()

        for cb in self._cleanup_callbacks:
            cb()

        self._cleanup_callbacks.clear()

    def _cleanup_sub_effects(self):
        for effect in self._sub_effects:
            effect.dispose()

        self._sub_effects.clear()

    def dispose(self):
        self._cleanup_sub_effects()
        self.cleanup_deps()
        # self.fn = None

    def _reset_age(self):
        self._age = 0

    def __repr__(self) -> str:
        return f"Effect(id ={self.id}, name={self.__debug_name})"



from signe.core.runtime import Executor, BatchExecutionScheduler
from signe.core.signal import Signal, SignalOption, TSignalOptionInitComp
from signe.core.effect import Effect
from signe.core.scope import Scope, IScope
from contextlib import contextmanager

from typing import (
    Any,
    Dict,
    List,
    TypeVar,
    Callable,
    Union,
    Sequence,
    overload,
    Optional,
    cast,
)


T = TypeVar("T")

exec = Executor()

TGetter = Callable[[], T]

TSetterParme = Union[T, Callable[[T], T]]
TSetter = Callable[[TSetterParme[T]], T]


class GlobalScopeManager:
    def __init__(self) -> None:
        self._stack: List[Scope] = []

    def _get_last_scope(self):
        idx = len(self._stack) - 1
        if idx < 0:
            return None

        return self._stack[idx]

    def new_scope(self):
        s = Scope()
        self._stack.append(s)
        return s

    def dispose_scope(self):
        s = self._get_last_scope()
        if s:
            s.dispose()
            self._stack.pop()

    def mark_effect_with_scope(self, scope: Optional[Scope], effect: Effect):
        s = scope
        if s:
            s.add_effect(effect)

        return effect

    def mark_effect(self, effect: Effect):
        return self.mark_effect_with_scope(self._get_last_scope(), effect)


_GLOBAL_SCOPE_MANAGER = GlobalScopeManager()


@contextmanager
def scope():
    _GLOBAL_SCOPE_MANAGER.new_scope()
    yield
    _GLOBAL_SCOPE_MANAGER.dispose_scope()


def createSignal(
    value: T,
    comp: TSignalOptionInitComp[T] = None,
    debug_name: Optional[str] = None,
):
    s = Signal(exec, value, SignalOption(comp), debug_name)

    return s.getValue, s.setValue


_TEffect_Fn = Callable[[Callable[..., T]], Effect[T]]


@overload
def effect(
    fn: None = ...,
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[IScope] = None,
) -> _TEffect_Fn[None]:
    ...


@overload
def effect(
    fn: Callable[..., None],
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[IScope] = None,
) -> Effect[None]:
    ...


def effect(
    fn: Optional[Callable[..., None]] = None,
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[IScope] = None,
) -> Union[_TEffect_Fn[None], Effect[None]]:
    kws = {
        "priority_level": priority_level,
        "debug_trigger": debug_trigger,
        "debug_name": debug_name,
    }

    if fn:
        res = Effect(exec, fn, **kws)
        if scope:
            scope.add_effect(res)
        else:
            _GLOBAL_SCOPE_MANAGER.mark_effect(res)
        return res
    else:

        def wrap(fn: Callable[..., None]):
            return effect(fn, **kws, scope=scope)

        return wrap


@overload
def computed(
    fn: None = ...,
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[IScope] = None,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    ...


@overload
def computed(
    fn: Callable[..., T],
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[IScope] = None,
) -> Callable[..., T]:
    ...


def computed(
    fn: Optional[Callable[[], T]] = None,
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[IScope] = None,
) -> Union[Callable[[Callable[..., T]], Callable[..., T]], Callable[..., T]]:
    kws = {
        "priority_level": priority_level,
        "debug_trigger": debug_trigger,
        "debug_name": debug_name,
    }

    if fn:
        current_effect = exec.effect_running_stack.get_current()

        def mark_scope(
            effect: Effect,
            global_scope=_GLOBAL_SCOPE_MANAGER._get_last_scope(),
        ):
            if scope:
                scope.add_effect(effect)
                return

            if global_scope:
                _GLOBAL_SCOPE_MANAGER.mark_effect(effect)

        def mark_sub_effect(
            effect: Effect,
            current_effect=current_effect,
        ):
            if current_effect is not None:
                current_effect._add_sub_effect(effect)

            del current_effect

        effect = None

        def wrap():
            nonlocal effect
            if effect is None:
                effect = Effect(exec, fn, **kws, capture_parent_effect=False)
                mark_scope(effect)
                mark_sub_effect(effect)

            return effect()

        return wrap

    else:

        def wrap_cp(fn: Callable[[], T]):
            return computed(fn, **kws, scope=scope)

        return wrap_cp


def batch(fn: Callable[[], None]):
    if isinstance(exec.current_execution_scheduler, BatchExecutionScheduler):
        fn()
        return

    batch_exec = BatchExecutionScheduler()

    try:
        exec.execution_scheduler_stack.set_current(batch_exec)
        fn()
        batch_exec.run_batch()
    finally:
        exec.execution_scheduler_stack.reset_current()


def cleanup(fn: Callable[[], None]):
    current_effect = exec.effect_running_stack.get_current()
    if current_effect:
        current_effect.add_cleanup_callback(fn)


def _getter_calls(fns: Sequence[TGetter[T]]):
    return tuple(fn() for fn in fns)


@overload
def on(
    getter: Union[TGetter[T], Sequence[TGetter[T]]],
    fn: Callable[..., None],
    onchanges=False,
    effect_kws: Optional[Dict[str, Any]] = None,
) -> Effect[None]:
    ...


@overload
def on(
    getter: Union[TGetter[T], Sequence[TGetter[T]]],
    fn: Optional[Callable[..., None]] = None,
    onchanges=False,
) -> Callable[[Callable], Effect[None]]:
    ...


def on(
    getter: Union[TGetter[T], Sequence[TGetter[T]]],
    fn: Optional[Callable[..., None]] = None,
    onchanges=False,
    effect_kws: Optional[Dict[str, Any]] = None,
):
    call_kws = {"onchanges": onchanges, "effect_kws": effect_kws}

    if fn is None:
        return cast(
            Callable[[Callable], Effect[None]],
            _on(getter, **call_kws),
        )

    return _on(getter, **call_kws)(fn)


import inspect
from dataclasses import dataclass


@dataclass(frozen=True)
class WatchedState:
    current: Any
    previous: Any


def _get_func_args_count(fn):
    return len(inspect.getfullargspec(fn).args)


# immediate
def _on(
    getter: Union[TGetter[T], Sequence[TGetter[T]]],
    onchanges=False,
    effect_kws: Optional[Dict[str, Any]] = None,
):
    getters = getter
    if not isinstance(getter, Sequence):
        getters = [getter]

    def getter_calls():
        return _getter_calls(getters)  # type: ignore

    def warp(fn: Callable[..., None]):
        args_count = _get_func_args_count(fn)
        prev_values = None

        def _on():
            nonlocal onchanges, prev_values
            current_values = getter_calls()  # type: ignore

            states = (
                WatchedState(cur, prev)
                for cur, prev in zip(current_values, prev_values or current_values)
            )

            prev_values = current_values

            if onchanges:
                onchanges = False
                return

            with pause_capture():
                if args_count == 0:
                    fn()
                else:
                    fn(*states)

        return effect(_on, **effect_kws or {})

    return warp


class pause_capture:
    def __init__(self) -> None:
        self.__cur = None

    def __enter__(self):
        if len(exec.effect_running_stack):
            self.__cur = exec.effect_running_stack.reset_current()

    def __exit__(self, *_):
        if self.__cur:
            exec.effect_running_stack.set_current(self.__cur)

        self.__cur = None



#!/usr/bin/env python
"""The setup script."""

from setuptools import setup, find_packages
import signe


with open("README.md", encoding="utf8") as readme_file:
    readme = readme_file.read()

requirements = ["typing_extensions"]

test_requirements = ["pytest>=3"]

setup(
    author="carson_jia",
    author_email="568166495@qq.com",
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
    ],
    description="...",
    entry_points={
        # 'console_scripts': [
        #     'test_prj=test_prj.cli:main',
        # ],
    },
    install_requires=requirements,
    license="MIT license",
    # long_description=readme,
    include_package_data=True,
    keywords=["signal", "signe", "S.js"],
    name="signe",
    packages=find_packages(include=["signe", "signe.*"]),
    data_files=[],
    test_suite="__tests",
    tests_require=test_requirements,
    url="",
    version=signe.__version__,
    zip_safe=False,
)



class NotPending:
    pass

    def __bool__(self):
        return False


NOT_PENDING = NotPending()


def is_not_pending(value):
    return isinstance(value, NotPending)


from enum import Enum


class EffectState(Enum):
    CURRENT = 0
    STALE = 1
    RUNNING = 2



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    NewType,
    TypeVar,
    Generic,
    Callable,
    Set,
    Union,
    Optional,
    cast,
)
from .consts import NOT_PENDING, is_not_pending


if TYPE_CHECKING:
    from .runtime import Executor
    from .effect import Effect

T = TypeVar("T")

TComp = TypeVar("TComp")

TSignalOptionComp = Callable[[TComp, TComp], bool]
TSignalOptionInitComp = Optional[Union[bool, TSignalOptionComp[TComp]]]


class SignalOption(Generic[T]):
    __slots__ = ("comp",)

    def __init__(self, comp: TSignalOptionInitComp[T] = None) -> None:
        if comp is None:
            comp = lambda old, new: old == new
        elif comp == False:
            comp = lambda old, new: False
        self.comp: TSignalOptionComp = comp  # type: ignore


class Signal(Generic[T]):
    _g_id = 0

    def __init__(
        self,
        executor: Executor,
        value: T,
        option: Optional[SignalOption[T]] = None,
        debug_name: Optional[str] = None,
    ) -> None:
        Signal._g_id += 1
        self.id = Signal._g_id
        self.__executor = executor
        self.value = value
        self.__dep_effects: Set[Effect] = set()
        self.option = option or SignalOption[T]()
        self._pending = NOT_PENDING
        self.__debug_name = debug_name

        self._option_comp = cast(Callable[[T, T], bool], self.option.comp)

    def getValue(self) -> T:
        current_effect = self.__executor.effect_running_stack.get_current()

        if current_effect:
            self.__dep_effects.add(current_effect)
            current_effect.add_dep_signal(self)

        return self.value  # type: ignore

    def setValue(self, value: Union[T, Callable[[T], T]]) -> T:
        if isinstance(value, Callable):
            value = value(self.value if is_not_pending(self._pending) else self._pending)  # type: ignore

        if self._option_comp(self.value, value):  # type: ignore
            return self.value  # type: ignore

        if len(self.__dep_effects) <= 0:
            self.value = value
            return self.value

        self._pending = value
        scheduler = self.__executor.current_execution_scheduler.add_signal(self)

        if not self.__executor.is_running:
            scheduler.run()

        return value  # type: ignore

    def update(self):
        self.value = self._pending
        self._pending = NOT_PENDING
        for sub in self.__dep_effects:
            sub._push_scheduler()

    def cleanup_dep_effect(self, effect: Effect):
        self.__dep_effects.remove(effect)

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, self.__class__):
            return self.__hash__() == __value.__hash__()

        return False

    def __repr__(self) -> str:
        return f"Signal(id= {self.id} , name = {self.__debug_name})"
